package com.qmth.exam.reserve.job;

import com.baomidou.mybatisplus.core.conditions.query.LambdaQueryWrapper;
import com.baomidou.mybatisplus.core.conditions.update.LambdaUpdateWrapper;
import com.qmth.boot.core.concurrent.service.ConcurrentService;
import com.qmth.exam.reserve.bean.apply.ApplyRecordCacheBean;
import com.qmth.exam.reserve.cache.CacheConstants;
import com.qmth.exam.reserve.cache.RedisClient;
import com.qmth.exam.reserve.entity.StudentApplyEntity;
import com.qmth.exam.reserve.service.StudentApplyService;
import org.redisson.api.RBlockingQueue;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.scheduling.annotation.Scheduled;
import org.springframework.stereotype.Component;

import java.util.concurrent.locks.Lock;

@Component
public class StudentApplyRecordJob {

    private static final Logger log = LoggerFactory.getLogger(StudentApplyRecordJob.class);

    @Autowired
    private StudentApplyService studentApplyService;

    @Autowired
    private ConcurrentService concurrentService;

    @Autowired
    private RedisClient redisClient;

    /**
     * 定时将“考生预约记录队列”中的数据保存至数据库
     * 注：每N秒执行一次
     */
    @Scheduled(fixedDelay = 30000, initialDelay = 30000)
    public void saveStudentApplyRecordJob() {
        long start = System.currentTimeMillis();

        try {
            Lock curLock = concurrentService.getLock(CacheConstants.LOCK_STUDENT_APPLY_RECORD);
            if (!curLock.tryLock()) {
                log.info("[JOB] locking...");
                return;
            }

            RBlockingQueue<ApplyRecordCacheBean> queue = redisClient.getRedissonClient()
                    .getBlockingQueue(CacheConstants.QUEUE_STUDENT_APPLY_RECORD);

            int queueSize = queue.size();
            log.info("[JOB] queue size:{}", queueSize);
            for (int i = 0; i < queueSize; i++) {
                ApplyRecordCacheBean value = queue.take();

                try {
                    this.saveOrUpdate(value);
                } catch (Exception e) {
                    // 保存至数据库失败，放回队列重试执行
                    boolean success = queue.offer(value);
                    log.error("[JOB] offerQueue:{} studentId:{} err:{}", success, value.getStudentId(), e.getMessage());
                }
            }

            curLock.unlock();
        } catch (Exception e) {
            log.error("[JOB] err:{}", e.getMessage(), e);
        } finally {
            long end = System.currentTimeMillis();
            log.info("[JOB] cost:{}ms", end - start);
        }
    }

    // @Transactional
    public void saveOrUpdate(ApplyRecordCacheBean bean) {
        LambdaQueryWrapper<StudentApplyEntity> wrapper = new LambdaQueryWrapper<>();
        wrapper.eq(StudentApplyEntity::getExamSiteId, bean.getExamSiteId());
        wrapper.eq(StudentApplyEntity::getTimePeriodId, bean.getTimePeriodId());
        wrapper.eq(StudentApplyEntity::getStudentId, bean.getStudentId());
        StudentApplyEntity existEntity = studentApplyService.getOne(wrapper);

        String msg = String.format("%s_%s_%s_%s", bean.getStudentId(), bean.getExamSiteId(),
                bean.getTimePeriodId(), bean.getCancel());
        if (existEntity != null) {
            if (!existEntity.getCancel()) {
                log.debug("{} 预约已存在，跳过！", msg);
                return;
            }

            // 存在“已取消”预约记录，则恢复预约
            LambdaUpdateWrapper<StudentApplyEntity> updateWrapper = new LambdaUpdateWrapper<>();
            updateWrapper.set(StudentApplyEntity::getCancel, bean.getCancel());
            updateWrapper.set(StudentApplyEntity::getOperateId, bean.getOperateId());
            updateWrapper.set(StudentApplyEntity::getUpdateTime, bean.getOperateTime());
            updateWrapper.eq(StudentApplyEntity::getId, existEntity.getId());
            studentApplyService.update(updateWrapper);
            log.debug("{} 预约修改！", msg);
        } else {
            // 不存在则新增预约记录
            StudentApplyEntity entity = new StudentApplyEntity();
            entity.setStudentId(bean.getStudentId());
            entity.setExamSiteId(bean.getExamSiteId());
            entity.setTimePeriodId(bean.getTimePeriodId());
            entity.setCancel(false);
            entity.setOperateId(bean.getOperateId());
            entity.setCreateTime(bean.getOperateTime());
            entity.setUpdateTime(bean.getOperateTime());
            studentApplyService.save(entity);
            log.debug("{} 预约新增！", msg);
        }
    }

}
